/*
**	Apple Macintosh Developer Technical Support
**
**	DirectoryCopy: A robust, general purpose directory copy routine.
**
**	by Jim Luther, Apple Developer Technical Support
**
**	File:		DirectoryCopy.c
**
**	Copyright  1992-1994 Apple Computer, Inc.
**	All rights reserved.
**
**	You may incorporate this sample code into your applications without
**	restriction, though the sample code has been provided "AS IS" and the
**	responsibility for its operation is 100% yours.  However, what you are
**	not permitted to do is to redistribute the source as "DSC Sample Code"
**	after having made changes. If you're going to re-distribute the source,
**	we require that you make it clear in the source that the code was
**	descended from Apple Sample Code, but that you've made changes.
*/

#ifndef __DIRECTORYCOPY__
#include "DirectoryCopy.h"
#endif


/*****************************************************************************/

/* local constants */

enum
{
	dirCopyBigCopyBuffSize  = 0x00004000,
	dirCopyMinCopyBuffSize  = 0x00000200
};


/*****************************************************************************/

/* local data structures */

/* The EnumerateGlobals structure is used to minimize the amount of
** stack space used when recursively calling CopyLevel and to hold
** global information that might be needed at any time. */
struct EnumerateGlobals
{
	Ptr			copyBuffer;			/* pointer to buffer used for file copy operations */
	long		bufferSize;			/* the size of the copy buffer */
	CopyErrProcPtr errorHandler;	/* pointer to error handling function */
	OSErr		error;				/* temporary holder of results - saves 2 bytes of stack each level */
	Boolean		bailout;			/* set to true to by error handling function if fatal error */
	short		destinationVRefNum;	/* the destination vRefNum */
	Str63		itemName;			/* the name of the current item */
	CInfoPBRec	myCPB;				/* the parameter block used for PBGetCatInfo calls */
};
typedef struct EnumerateGlobals EnumerateGlobals;
typedef EnumerateGlobals *EnumerateGlobalsPtr;


/* The PreflightGlobals structure is used to minimize the amount of
** stack space used when recursively calling GetLevelSize and to hold
** global information that might be needed at any time. */
struct PreflightGlobals
{
	OSErr		result;				/* temporary holder of results - saves 2 bytes of stack each level */
	Str63		itemName;			/* the name of the current item */
	CInfoPBRec	myCPB;				/* the parameter block used for PBGetCatInfo calls */

	long		dstBlksPerAllocBlk;	/* the number of 512 byte blocks per allocation block on destination */
	long		allocBlksNeeded;	/* the total number of allocation blocks needed  */

	long 		tempBlocks;			/* temporary storage for calculations (save some stack space)  */
};
typedef struct PreflightGlobals PreflightGlobals;
typedef PreflightGlobals *PreflightGlobalsPtr;

/*****************************************************************************/

/* static prototypes */

static	void	GetLevelSize(long currentDirID,
							 PreflightGlobalsPtr theGlobals);

static	OSErr	PreflightDirectoryCopySpace(short srcVRefNum,
											long srcDirID,
											short dstVRefNum,
											Boolean *spaceOK);

static	void	CopyLevel(long sourceDirID,
						  long dstDirID,
						  EnumerateGlobalsPtr theGlobals);
						  
/*****************************************************************************/

static	void	GetLevelSize(long currentDirID,
							 PreflightGlobalsPtr theGlobals)
{
	short	index = 1;    
	do {
          theGlobals->myCPB.dirInfo.ioFDirIndex = index;
          theGlobals->myCPB.dirInfo.ioDrDirID = currentDirID;	/* we need to do this every time */
          														/* through, since GetCatInfo  */
          														/* returns ioFlNum in this field */
          theGlobals->result = PBGetCatInfoSync((CInfoPBPtr)&theGlobals->myCPB);
          if (theGlobals->result == noErr) {
             if ((theGlobals->myCPB.dirInfo.ioFlAttrib & ioDirMask) != 0) {
				/* we have a directory */
    
				GetLevelSize(theGlobals->myCPB.dirInfo.ioDrDirID, theGlobals); /* recurse */
				theGlobals->result = noErr; /* clear error return on way back */
			}
			else {
				/* we have a file - add its allocation blocks to allocBlksNeeded */
    				
				/* get number of 512-byte blocks needed for data fork */
				theGlobals->tempBlocks = (theGlobals->myCPB.hFileInfo.ioFlLgLen % 512) ?
											((theGlobals->myCPB.hFileInfo.ioFlLgLen >> 9) + 1) :
											(theGlobals->myCPB.hFileInfo.ioFlLgLen >> 9);
				/* now, calculate number of new allocation blocks needed */
				theGlobals->allocBlksNeeded += (theGlobals->tempBlocks % theGlobals->dstBlksPerAllocBlk) ?
												((theGlobals->tempBlocks / theGlobals->dstBlksPerAllocBlk) + 1) :
												(theGlobals->tempBlocks / theGlobals->dstBlksPerAllocBlk);
			
				/* get number of 512-byte blocks needed for resource fork */
				theGlobals->tempBlocks = (theGlobals->myCPB.hFileInfo.ioFlRLgLen % 512) ?
											((theGlobals->myCPB.hFileInfo.ioFlRLgLen >> 9) + 1) :
											(theGlobals->myCPB.hFileInfo.ioFlRLgLen >> 9);
				/* now, calculate number of new allocation blocks needed */
				theGlobals->allocBlksNeeded += (theGlobals->tempBlocks % theGlobals->dstBlksPerAllocBlk) ?
												((theGlobals->tempBlocks / theGlobals->dstBlksPerAllocBlk) + 1) :
												(theGlobals->tempBlocks / theGlobals->dstBlksPerAllocBlk);
			}
		}
		++index;
	} while (theGlobals->result == noErr);
}

/*****************************************************************************/

static	OSErr	PreflightDirectoryCopySpace(short srcVRefNum,
											long srcDirID,
											short dstVRefNum,
											Boolean *spaceOK)
{
	HParamBlockRec pb;
	OSErr error;
	long dstFreeBlocks;
	PreflightGlobals theGlobals;
	
	/* Get the number of 512 byte blocks per allocation block and */
	/* number of free allocation blocks on the destination volume */
	pb.volumeParam.ioVRefNum = dstVRefNum;
	pb.volumeParam.ioNamePtr = nil;
	pb.volumeParam.ioVolIndex = 0;		/* use ioVRefNum only */
	error = PBHGetVInfoSync(&pb);
	if (error == noErr)
	{
		dstFreeBlocks = pb.volumeParam.ioVFrBlk;
		
		/* get allocation block size (always multiple of 512) and divide by 512
		  to get number of 512-byte blocks per allocation block */
		theGlobals.dstBlksPerAllocBlk = (pb.volumeParam.ioVAlBlkSiz >> 9);
		
		theGlobals.allocBlksNeeded = 0;

		theGlobals.myCPB.dirInfo.ioNamePtr = theGlobals.itemName;
		theGlobals.myCPB.dirInfo.ioVRefNum = srcVRefNum;
		
		
		GetLevelSize(srcDirID, &theGlobals);
		
		/* Is there enough room on the destination volume for the source file? */
		*spaceOK = (theGlobals.allocBlksNeeded <= dstFreeBlocks);
	}
	return (error);
}

/*****************************************************************************/

static	void	CopyLevel(long sourceDirID,
						  long dstDirID,
						  EnumerateGlobalsPtr theGlobals)
{
	long currentSrcDirID;
	long newDirID;
	short index = 1;
	
	do
	{	/* Isn't C great... What I'd give for a "WITH theGlobals DO" about now... */
	
		/* Get next source item at the current directory level */
		
		theGlobals->myCPB.dirInfo.ioFDirIndex = index;
		theGlobals->myCPB.dirInfo.ioDrDirID = sourceDirID;
		theGlobals->error = PBGetCatInfoSync((CInfoPBPtr)&theGlobals->myCPB);		

		if (theGlobals->error == noErr) {
		
			/* We have an item.  Is it a file or directory? */
			if (theGlobals->myCPB.hFileInfo.ioFlAttrib & 0x10) {
			
				/* we have a directory */
				
				/* create a new directory at the destination. No errors allowed! */
				theGlobals->error = DirCreate(theGlobals->destinationVRefNum, dstDirID, theGlobals->itemName, &newDirID);
				
				if (theGlobals->error == noErr) {
				
					/* Save the current source directory ID where we can get it when we come back
					** from recursion land. */
					
					currentSrcDirID = theGlobals->myCPB.dirInfo.ioDrDirID;
					
					/* Dive again (copy the directory level we just found below this one) */
					
					CopyLevel(theGlobals->myCPB.dirInfo.ioDrDirID, newDirID, theGlobals);
					
					if (!theGlobals->bailout) {
					
						/* Copy comment from old to new directory. */
						
						theGlobals->error = CopyComment(theGlobals->myCPB.dirInfo.ioVRefNum, currentSrcDirID, nil, theGlobals->destinationVRefNum, newDirID, nil);
						
						/* handle any errors from CopyComment */
						if ((theGlobals->error != noErr) && (theGlobals->errorHandler != nil))
							theGlobals->bailout =  (* (theGlobals->errorHandler)) (theGlobals->error, copyDirCommentOp,
																				theGlobals->myCPB.dirInfo.ioVRefNum, currentSrcDirID, nil,
																				theGlobals->destinationVRefNum, newDirID, nil);
					}

					if (!theGlobals->bailout) {
					
						/* Copy directory attributes (dates, etc.) to newDirID. */
						
						theGlobals->error = CopyFileMgrAttributes(theGlobals->myCPB.dirInfo.ioVRefNum, currentSrcDirID, nil, theGlobals->destinationVRefNum, newDirID, nil, true);
						
						/* handle any errors from CopyFileMgrAttributes */
						if ((theGlobals->error != noErr) && (theGlobals->errorHandler != nil))
							theGlobals->bailout =  (* (theGlobals->errorHandler)) (theGlobals->error, copyDirFMAttributesOp,
																				theGlobals->myCPB.dirInfo.ioVRefNum, currentSrcDirID, nil,
																				theGlobals->destinationVRefNum, newDirID, nil);
					}
				}
				else	/* error handling for DirCreate */
					if (theGlobals->errorHandler != nil)
						theGlobals->bailout = (* (theGlobals->errorHandler)) (theGlobals->error, dirCreateOp,
													theGlobals->myCPB.dirInfo.ioVRefNum, currentSrcDirID, nil,
													theGlobals->destinationVRefNum, dstDirID, theGlobals->itemName);
				theGlobals->error = noErr;	/* clear error return on way back */
			}
			else {
			
				/* We have a file, so copy it */
				
				theGlobals->error = FileCopy(theGlobals->myCPB.hFileInfo.ioVRefNum,
											 theGlobals->myCPB.hFileInfo.ioFlParID,
											 theGlobals->itemName,
											 theGlobals->destinationVRefNum,
											 dstDirID,
											 nil,
											 nil,
											 theGlobals->copyBuffer,
											 theGlobals->bufferSize,
											 false);
						
				/* handle any errors from FileCopy */
				if ((theGlobals->error != noErr) && (theGlobals->errorHandler != nil))
					theGlobals->bailout =  (* (theGlobals->errorHandler)) (theGlobals->error, fileCopyOp,
																		theGlobals->myCPB.hFileInfo.ioVRefNum, theGlobals->myCPB.hFileInfo.ioFlParID, theGlobals->itemName,
																		theGlobals->destinationVRefNum, dstDirID, nil);
			}
		}
		else {	/* error handling for PBGetCatInfo */
			/* it's normal to get a fnfErr when indexing; that only means you've hit the end of the directory */
			if ((theGlobals->error != fnfErr) && (theGlobals->errorHandler != nil)) { 
				theGlobals->bailout = (* (theGlobals->errorHandler)) (theGlobals->error, getNextItemOp, theGlobals->myCPB.dirInfo.ioVRefNum, sourceDirID, nil, 0, 0, nil);
			}
		}
		++index; /* prepare to get next item */
	}
	while ((theGlobals->error == noErr) && (!theGlobals->bailout)); /* time to fall back a level? */
}

/*****************************************************************************/

pascal	OSErr	DirectoryCopy(short srcVRefNum,
							  long srcDirID,
							  StringPtr srcName,
							  short dstVRefNum,
							  long dstDirID,
							  StringPtr dstName,
							  Ptr copyBufferPtr,
							  long copyBufferSize,
							  Boolean preflight,
							  CopyErrProcPtr copyErrHandler)
{
	EnumerateGlobals theGlobals;
	Boolean	isDirectory;
	OSErr	error;
	Boolean ourCopyBuffer = false;
	Str63	srcDirName;
	Boolean spaceOK;
	
	/* Make sure a copy buffer is allocated. */
	if (copyBufferPtr == nil) {
		/* The caller didn't supply a copy buffer so grab one from the application heap.
		** Try to get a big copy buffer, if we can't, try for a 512-byte buffer.
		** If 512 bytes aren't available, we're in trouble. */
		copyBufferSize = dirCopyBigCopyBuffSize;
		copyBufferPtr = NewPtr(copyBufferSize);
		if (copyBufferPtr == nil) {
			copyBufferSize = dirCopyMinCopyBuffSize;
			copyBufferPtr = NewPtr(copyBufferSize);
			if (copyBufferPtr == nil)
				return(MemError());
		}
		ourCopyBuffer = true;
	}
	
	/* Get the real dirID where we're copying from and make sure it is a directory. */
	error = GetDirID(srcVRefNum, srcDirID, srcName, &srcDirID, &isDirectory);
	if (error != noErr)
		goto ErrorExit;
	if (!isDirectory) {
		error =  dirNFErr;
		goto ErrorExit;
	}
		
	/*  Get the real dirID where we're going to put the copy and make sure it is a directory. */
	error = GetDirID(dstVRefNum, dstDirID, dstName, &dstDirID, &isDirectory);
	if (error != noErr)
		goto ErrorExit;
	if (!isDirectory) {
		error =  dirNFErr;
		goto ErrorExit;
	}
	
	/* Get the real vRefNum of both the source and destination */
	error = DetermineVRefNum(srcName, srcVRefNum, &srcVRefNum);
	if (error != noErr)
		goto ErrorExit;
	error = DetermineVRefNum(dstName, dstVRefNum, &dstVRefNum);
	if (error != noErr)
		goto ErrorExit;
	
	if (preflight) {
		error = PreflightDirectoryCopySpace(srcVRefNum, srcDirID, dstVRefNum, &spaceOK);
		if (error != noErr)
			goto ErrorExit;
		if (!spaceOK) {
			error = dskFulErr; /* not enough room on destination */
			goto ErrorExit;
		}
	}

	/* Create the new directory in the destination directory with the */
	/* same name as the source directory. */
	error = GetDirName(srcVRefNum, srcDirID, srcDirName);
	if (error != noErr)
		goto ErrorExit;
	error = DirCreate(dstVRefNum, dstDirID, srcDirName, &dstDirID);
	if (error != noErr)
		goto ErrorExit;
	
	/* dstDirID is now the newly created directory! */
		
	/* Set up the globals we need to access from the recursive routine. */
	theGlobals.copyBuffer = copyBufferPtr;
	theGlobals.bufferSize = copyBufferSize;
	theGlobals.destinationVRefNum = dstVRefNum; /* so we can get to it always */
	theGlobals.myCPB.hFileInfo.ioNamePtr = (StringPtr)&theGlobals.itemName;
	theGlobals.myCPB.hFileInfo.ioVRefNum = srcVRefNum;
	theGlobals.errorHandler = copyErrHandler;
	theGlobals.bailout = false;
		
	/* Here we go into recursion land... */
	CopyLevel(srcDirID, dstDirID, &theGlobals);
	
	if (!theGlobals.bailout) {
	
		/* Copy comment from source to destination directory. */
		
		error = CopyComment(srcVRefNum, srcDirID, nil, dstVRefNum, dstDirID, nil);
		
		/* handle any errors from CopyComment */
		if ((error != noErr) && (copyErrHandler != nil))
			theGlobals.bailout =  (* copyErrHandler) (error, copyDirCommentOp,
									srcVRefNum, srcDirID, nil,
									dstVRefNum, dstDirID, nil);
	}

	if (!theGlobals.bailout) {
	
		/* Copy the File Manager attributes */
		
		error = CopyFileMgrAttributes(srcVRefNum, srcDirID, nil,
					dstVRefNum, dstDirID, nil, true);
		
		/* handle any errors from CopyFileMgrAttributes */
		if ((error != noErr) && (copyErrHandler != nil))
			theGlobals.bailout = (* copyErrHandler) (error, copyDirFMAttributesOp,
									srcVRefNum, srcDirID, nil,
									dstVRefNum, dstDirID, nil);
	}

ErrorExit:
	/* Get rid of the copy buffer if we allocated it. */
	if (ourCopyBuffer)
		DisposPtr(copyBufferPtr);

	return (error);
}

/*****************************************************************************/

pascal	OSErr	FSpDirectoryCopy(const FSSpec *srcSpec,
								 const FSSpec *dstSpec,
								 Ptr copyBufferPtr,
								 long copyBufferSize,
								 Boolean preflight,
								 CopyErrProcPtr copyErrHandler)
{
	return (DirectoryCopy(srcSpec->vRefNum, srcSpec->parID, (StringPtr)srcSpec->name,
						  dstSpec->vRefNum, dstSpec->parID, (StringPtr)dstSpec->name,
						  copyBufferPtr, copyBufferSize, preflight, copyErrHandler));
}

/*****************************************************************************/

