/*
**	Apple Macintosh Developer Technical Support
**
**	FileCopy: A robust, general purpose file copy routine.
**
**	by Jim Luther, Apple Developer Technical Support
**
**	File:		FileCopy.c
**
**	Copyright  1992-1994 Apple Computer, Inc.
**	All rights reserved.
**
**	You may incorporate this sample code into your applications without
**	restriction, though the sample code has been provided "AS IS" and the
**	responsibility for its operation is 100% yours.  However, what you are
**	not permitted to do is to redistribute the source as "DSC Sample Code"
**	after having made changes. If you're going to re-distribute the source,
**	we require that you make it clear in the source that the code was
**	descended from Apple Sample Code, but that you've made changes.
*/

#ifndef __FILECOPY__
#include "FileCopy.h"
#endif


/*****************************************************************************/

/* local constants */

/*	The deny-mode privileges to use when opening the source and destination files. */

enum
{
	srcCopyMode = dmRdDenyWr,
	dstCopyMode = dmWrDenyRdWr
};

/*	The largest (16K) and smallest (.5K) copy buffer to use if the caller doesn't supply 
**	their own copy buffer. */

enum
{
	bigCopyBuffSize  = 0x00004000,
	minCopyBuffSize  = 0x00000200
};

/*****************************************************************************/

/* static prototypes */

static	void	pcpy(StringPtr d,
					 StringPtr s);

static	OSErr	PreflightFileCopySpace(short srcVRefNum,
									   long srcDirID,
									   const Str255 srcName,
									   StringPtr dstVolName,
									   short dstVRefNum,
									   Boolean *spaceOK);
/*	PreflightFileCopySpace determines if there's enough space on a
	volume to copy the specified file to that volume.
	Note: The results of this routine are not perfect. For example if the
	volume's catalog or extents overflow file grows when the new file is
	created, more allocation blocks may be needed beyond those needed for
	the file's data and resource forks.

	srcVRefNum		input:	Volume specification of the file's current
							location.
	srcDirID		input:	Directory ID of the file's current location.
	srcName			input:	The name of the file.
	dstVolName		input:	A pointer to the name of the volume where
							the file will be copied or nil.
	dstVRefNum		input:	Volume specification indicating the volume
							where the file will be copied.
	spaceOK			output:	true if there's enough space on the volume for
							the file's data and resource forks.
*/

/*****************************************************************************/

/* Copy a Pascal-string. */
static	void	pcpy(StringPtr d,
					 StringPtr s)
{
	short	i;

	i = *s;
	do {
		d[i] = s[i];
	} while (i--);
}

/*****************************************************************************/

static	OSErr	PreflightFileCopySpace(short srcVRefNum,
									   long srcDirID,
									   const Str255 srcName,
									   StringPtr dstVolName,
									   short dstVRefNum,
									   Boolean *spaceOK)
{
	HParamBlockRec pb;
	OSErr error;
	Str255 tempPathname;
	long dstFreeBlocks;
	long dstBlksPerAllocBlk;
	long srcDataBlks;
	long srcResourceBlks;
	
	/* Get the number of 512 byte blocks per allocation block and */
	/* number of free allocation blocks on the destination volume */
	pb.volumeParam.ioVRefNum = dstVRefNum;
	if (dstVolName == nil) {
		pb.volumeParam.ioNamePtr = nil;
		pb.volumeParam.ioVolIndex = 0;		/* use ioVRefNum only */
	}
	else {
		pcpy((StringPtr)tempPathname, dstVolName);	/* make a copy of the string and */
		pb.volumeParam.ioNamePtr = (StringPtr)tempPathname;	/* use the copy so original isn't trashed */
		pb.volumeParam.ioVolIndex = -1;	/* use ioNamePtr/ioVRefNum combination */
	}
	error = PBHGetVInfoSync(&pb);
	if (error == noErr)
	{
		/* get allocation block size (always multiple of 512) and divide by 512
		  to get number of 512-byte blocks per allocation block */
		dstBlksPerAllocBlk = (pb.volumeParam.ioVAlBlkSiz >> 9);
		dstFreeBlocks = pb.volumeParam.ioVFrBlk;
		
		/* Now, get the size of the file's data resource forks */
		pb.fileParam.ioNamePtr = (StringPtr)srcName;
		pb.fileParam.ioVRefNum = srcVRefNum;
		pb.fileParam.ioDirID = srcDirID;
		pb.fileParam.ioFDirIndex = 0;
		error = PBHGetFInfoSync(&pb);
		if (error == noErr)
		{
			/* get number of 512-byte blocks needed for data fork */
			srcDataBlks = (pb.fileParam.ioFlLgLen % 512) ? ((pb.fileParam.ioFlLgLen >> 9) + 1) : (pb.fileParam.ioFlLgLen >> 9);
			/* now, calculate number of new allocation blocks needed */
			srcDataBlks = (srcDataBlks % dstBlksPerAllocBlk) ? ((srcDataBlks / dstBlksPerAllocBlk) + 1) : (srcDataBlks / dstBlksPerAllocBlk);
		
			/* get number of 512-byte blocks needed for resource fork */
			srcResourceBlks = (pb.fileParam.ioFlRLgLen % 512) ? ((pb.fileParam.ioFlRLgLen >> 9) + 1) : (pb.fileParam.ioFlRLgLen >> 9);
			/* now, calculate number of new allocation blocks needed */
			srcResourceBlks = (srcResourceBlks % dstBlksPerAllocBlk) ? ((srcResourceBlks / dstBlksPerAllocBlk) + 1) : (srcResourceBlks / dstBlksPerAllocBlk);
			
			/* Is there enough room on the destination volume for the source file? */
			*spaceOK = ((srcDataBlks + srcResourceBlks) <= dstFreeBlocks);
		}
	}
	return (error);
}

/*****************************************************************************/

pascal	OSErr	FileCopy(short srcVRefNum,
						 long srcDirID,
						 const Str255 srcName,
						 short dstVRefNum,
						 long dstDirID,
						 StringPtr dstPathname,
						 StringPtr copyName,
						 Ptr copyBufferPtr,
						 long copyBufferSize,
						 Boolean preflight)
{
	OSErr	err;

	short	srcRefNum = 0,			/* 0 when source data and resource fork are closed  */
			dstDataRefNum = 0,		/* 0 when destination data fork is closed */
			dstRsrcRefNum = 0;		/* 0 when destination resource fork is closed */
	
	Str63	dstName;				/* The filename of the destination. It might be the
									** source filename, it might be a new name... */
	
	GetVolParmsInfoBuffer infoBuffer; /* Where PBGetVolParms dumps its info */
	long	srcServerAdr;			/* AppleTalk server address of source (if any) */
	
	Boolean	dstCreated = false,		/* true when destination file has been created */
			ourCopyBuffer = false,	/* true if we had to allocate the copy buffer */
			isDirectory;			/* true if destination is really a directory */
	
	long	tempLong;
	short	tempInt;
	
	Boolean	spaceOK;				/* true if there's enough room to copy the file to the destination volume */

	/* Preflight for size */
	if (preflight) {
		err = PreflightFileCopySpace(srcVRefNum, srcDirID, srcName,
									 dstPathname, dstVRefNum, &spaceOK);
		if (err != noErr)
			return(err);
		if (!spaceOK)
			return(dskFulErr);
	}

	/* get the destination's real dirID and make sure it really is a directory */
	err = GetDirID(dstVRefNum, dstDirID, dstPathname, &dstDirID, &isDirectory);
	if (err != noErr)
		goto ErrorExit;
	if (!isDirectory)
		return (dirNFErr);

	/* get the destination's real vRefNum */
	err = DetermineVRefNum(dstPathname, dstVRefNum, &dstVRefNum);
	if (err != noErr)
		goto ErrorExit;
	
	/* See if PBHCopyFile can be used.  Using PBHCopyFile saves time by letting the file server
	** copy the file if the source and destination locations are on the same file server. */
	tempLong = sizeof(infoBuffer);
	err = HGetVolParms((StringPtr)srcName, srcVRefNum, &infoBuffer, &tempLong);
	if ((err != noErr) && (err != paramErr))
		return(err);

	if ((err != paramErr) && hasCopyFile(infoBuffer)) {
		/* The source volume supports PBHCopyFile. */
		srcServerAdr = infoBuffer.vMServerAdr;

		/* Now, see if the destination volume is on the same file server. */
		tempLong = sizeof(infoBuffer);
		err = HGetVolParms(nil, dstVRefNum, &infoBuffer, &tempLong);
		if ((err != noErr) && (err != paramErr))
			return(err);
		if ((err != paramErr) && (srcServerAdr == infoBuffer.vMServerAdr)) {
			/* Source and Dest are on same server and PBHCopyFile is supported. Copy with CopyFile. */
			err = HCopyFile(srcVRefNum, srcDirID, srcName, dstVRefNum, dstDirID, nil, copyName);
			if (err != noErr)
				goto ErrorExit;
			/* AppleShare's CopyFile clears the isAlias bit, so I still need to attempt to copy
			   the File's attributes to attempt to get things right. */
			return(CopyFileMgrAttributes(srcVRefNum, srcDirID, (StringPtr)srcName,
										 dstVRefNum, dstDirID, (StringPtr)&copyName, true));
		}
	}

	/* If we're here, then PBHCopyFile couldn't be used so we have to copy the file by hand. */

	/* Make sure a copy buffer is allocated. */
	if (copyBufferPtr == nil) {
		/* The caller didn't supply a copy buffer so grab one from the application heap.
		** Try to get a big copy buffer, if we can't, try for a 512-byte buffer.
		** If 512 bytes aren't available, we're in trouble. */
		copyBufferSize = bigCopyBuffSize;
		copyBufferPtr = NewPtr(copyBufferSize);
		if (copyBufferPtr == nil) {
			copyBufferSize = minCopyBuffSize;
			copyBufferPtr = NewPtr(copyBufferSize);
			if (copyBufferPtr == nil)
				return(MemError());
		}
		ourCopyBuffer = true;
	}

	/* Open the source data fork. */
	err = HOpenAware(srcVRefNum, srcDirID, srcName, srcCopyMode, &srcRefNum);
	if (err != noErr)
		goto ErrorExit;
	
	/* See if the copy will be renamed. */
	if (copyName != nil)				/* Did caller supply copy file name? */
		pcpy((StringPtr)&dstName, copyName);	/* Yes, use the caller supplied copy file name. */
	else {								/* They didn't, so get the source file name and use it. */
		err = GetFileLocation(srcRefNum, &tempInt, &tempLong, (StringPtr)&dstName);
		if (err != noErr)
			goto ErrorExit;
	}

	/* Create the destination file. */
	err = HCreateMinimum(dstVRefNum, dstDirID, dstName);
	if (err != noErr)
		goto ErrorExit;
	dstCreated = true;	/* After creating the destination file, any
						** error conditions should delete the destination file */

	/* An AppleShare dropbox folder is a folder for which the user has the Make Changes
	** privilege (write access), but not See Files (read access) and See Folders (search access).
	** Copying a file into an AppleShare dropbox presents some special problems. Here are the
	** rules we have to follow to copy a file into a dropbox:
	**  File attributes can be changed only when both forks of a file are empty.
	**  DeskTop Manager comments can be added to a file only when both forks of a file 
	**   are empty.
	**  A fork can be opened for write access only when both forks of a file are empty.
	** So, with those rules to live with, we'll do those operations now while both forks
	** are empty. */

	/* Copy attributes but don't lock the destination. */
	err = CopyFileMgrAttributes(srcVRefNum, srcDirID, (StringPtr)srcName,
								dstVRefNum, dstDirID, (StringPtr)&dstName, false);
	if (err != noErr)
		goto ErrorExit;

	/* Attempt to copy the comments while both forks are empty.
	** Ignore the result because we really don't care if it worked or not. */
	CopyComment(srcVRefNum, srcDirID, (StringPtr)srcName, dstVRefNum, dstDirID, (StringPtr)&dstName);

	/* Open the destination data fork. */
	err = HOpenAware(dstVRefNum, dstDirID, dstName, dstCopyMode, &dstDataRefNum);
	if (err != noErr)
		goto ErrorExit;

	/* Open the destination resource fork. */
	err = HOpenRFAware(dstVRefNum, dstDirID, dstName, dstCopyMode, &dstRsrcRefNum);
	if (err != noErr)
		goto ErrorExit;

	/* Copy the data fork. */
	err = CopyFork(srcRefNum, dstDataRefNum, copyBufferPtr, copyBufferSize);
	if (err != noErr)
		goto ErrorExit;

	/* Close both data forks and clear reference numbers. */
	FSClose(srcRefNum);
	FSClose(dstDataRefNum);
	srcRefNum = dstDataRefNum = 0;

	/* Open the source resource fork. */
	err = HOpenRFAware(srcVRefNum, srcDirID, srcName, srcCopyMode, &srcRefNum);
	if (err != noErr)
		goto ErrorExit;

	/* Copy the resource fork. */
	err = CopyFork(srcRefNum, dstRsrcRefNum, copyBufferPtr, copyBufferSize);
	if (err != noErr)
		goto ErrorExit;

	/* Close both resource forks and clear reference numbers. */
	FSClose(srcRefNum);
	FSClose(dstRsrcRefNum);
	srcRefNum = dstRsrcRefNum = 0;

	/* Get rid of the copy buffer if we allocated it. */
	if (ourCopyBuffer)
		DisposPtr(copyBufferPtr);

	/* Attempt to copy attributes again to set mod date.  Copy lock condition this time
	** since we're done with the copy operation.  This operation will fail if we're copying
	** into an AppleShare dropbox, so we don't check for error conditions. */
	CopyFileMgrAttributes(srcVRefNum, srcDirID, (StringPtr)srcName,
							dstVRefNum, dstDirID, (StringPtr)&dstName, true);

	/* Hey, we did it! */
	return (noErr);
	
ErrorExit:
	if (srcRefNum)
		FSClose(srcRefNum);		/* Close the source file */
	if (dstDataRefNum)
		FSClose(dstDataRefNum);	/* Close the destination file data fork */
	if (dstRsrcRefNum)
		FSClose(dstRsrcRefNum);	/* Close the destination file resource fork */
	if (dstCreated)
		HDelete(dstVRefNum, dstDirID, dstName);	/* Delete dest file.  This may fail if the file 
												   is in a "drop folder" */
	if (ourCopyBuffer)			/* dispose of any memory we allocated */
		DisposPtr(copyBufferPtr);
	return (err);
}

/*****************************************************************************/

pascal	OSErr	FSpFileCopy(const FSSpec *srcSpec,
							const FSSpec *dstSpec,
							StringPtr copyName,
							Ptr copyBufferPtr,
							long copyBufferSize,
							Boolean preflight)
{
	return (FileCopy(srcSpec->vRefNum, srcSpec->parID, srcSpec->name,
					 dstSpec->vRefNum, dstSpec->parID, (StringPtr)dstSpec->name,
					 copyName, copyBufferPtr, copyBufferSize, preflight));
}

/*****************************************************************************/

